
/* ***********************************************************************
/* File is distrubuted under GPL license, see LICENSE file for datails. /*
/* Author: Lukasz Wojtow <lw@ftw.zamosc.pl>				/*
/* Interesting ideas: Rafal Wijata <www.wijata.com> 			/*
/* Dedicate to: Creed							/*
/* Last change: 16.09.2002						/*
/************************************************************************/

#define CORE_PRIVATE
#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_main.h"
#include "http_protocol.h"
#include "util_script.h"
 
#include <stdio.h>
#include <pwd.h>
#include <grp.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <limits.h>
#include <stdlib.h>
 
#include "config.h"
#include "version.h"

#define NOTHING -1      // do nothing with param
#define DYNAMIC -2	// dynamic config, allowed only for ~user requests
#define OK 0
#define ERR 1

static char *last_server_name=NULL;	//per child ?
 
#if ISDEBUG==1
//warning at format string, use this with care
#define DEBUG(MSG...) ap_log_rerror(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,r,MSG)
#endif
 
#if ISDEBUG==0
#define DEBUG(MSG...)
#endif

#define LOG_ERROR(MSG...) ap_log_rerror(APLOG_MARK,APLOG_EMERG,r,MSG)
#define LOG_NOERROR(MSG...) ap_log_rerror(APLOG_MARK,APLOG_NOERRNO,r,MSG)
#define ESS 255 // error string size

module MODULE_VAR_EXPORT diffprivs_module;

typedef struct t_control {
   unsigned int user;       // do ~user privs?
   unsigned int didall;     // did all or not?
   unsigned int setgroups;  // call setgroup(0,NULL)?
   unsigned int pres;	    // is any directive present or not
} t_control;

typedef struct t_user {
   uid_t    uid;	    // uid for setuid() // ~user 
   gid_t    gid;	    // gid for setgid() // ~user
   char *chrootdir;         // dir for chroot   // ~user
   char *relative;          // dir for relative chroot // ~user
} t_user;

typedef struct t_host {
   uid_t uid;          	    // uid for setuid()
   gid_t gid;		    // gid for setgid()
   char *chrootdir;         // dir for chroot
} t_host;

// configuration record for module
typedef struct t_cfg
{
   t_host    *host;	    // struct with host requests config
   t_user    *user;         // struct with user requests config
   t_control *control;      // struct with control parameters
} t_cfg;

// add info to server version
static void mod_diffprivs_init(server_rec *s,pool *p)
{
   ap_add_version_component("diffprivs/"VERSION);
   return;
};

// create place for (this virtual) server config
static void *mod_diffprivs_server_config(pool *p, server_rec *s)
{
   t_cfg *cfg= (t_cfg *) ap_palloc(p,sizeof(t_cfg));
   cfg->control= (t_control *) ap_palloc(p,sizeof(t_control));
   cfg->host= (t_host *) ap_palloc(p,sizeof(t_host));
   cfg->user= (t_user *) ap_palloc(p,sizeof(t_user));

   cfg->control->didall=0;
   cfg->control->pres=0;
   cfg->control->user=0;
   cfg->control->setgroups=0;
   cfg->user->uid=(uid_t)NOTHING;
   cfg->user->gid=(gid_t)NOTHING;
   cfg->user->chrootdir=NULL;
   cfg->user->relative=NULL;
   cfg->host->uid=(uid_t)NOTHING;
   cfg->host->gid=(gid_t)NOTHING;
   cfg->host->chrootdir=NULL;
   return (void *) cfg;
}

static int try_setgroups(request_rec *r)
{
	if(setgroups(0,NULL)==-1) {
		LOG_ERROR("euid %u cannot call setgroups()\n",geteuid());
		return ERR;
	}
	return OK;
}

static int try_setuid(request_rec *r, uid_t uid)
{
	if(uid==NOTHING) return OK;
	if(uid<MIN_UID) {
		LOG_NOERROR("uid %u is less than minimal allowed uid %d\n",
				uid,MIN_UID);
		return ERR;
	}
	if(setuid(uid)!=0) {
		LOG_ERROR("euid %u cannot call setuid(%d)\n",geteuid(),uid);
		return ERR;
	}
	return OK;
}

static int try_setgid(request_rec *r, gid_t gid)
{
	if(gid==NOTHING) return OK;
	if(gid<MIN_GID) {
		LOG_NOERROR("gid %u is less than minimal allowed gid %d\n",
				gid,MIN_GID);
		return ERR;
	}
	if(setgid(gid)!=0) {
		LOG_ERROR("euid %u cannot call setgid(%d)\n",geteuid(),gid);
		return ERR;
	}
	return OK;
}

static int try_chroot(request_rec *r, char *dir)
{
	DEBUG("try_chroot() to dir %s",dir);
	// checks was already done while reading config.
	// if dynamic dir is used, it will match criteria
	if(chroot(dir)!=0) {
		LOG_ERROR("euid %u cannot call chroot(%s)\n",geteuid(),dir);
		return ERR;
	}
	if(chdir("/")!=0) {
		LOG_ERROR(
		"euid %d cannot call chdir(/) after chroot(%s)",
		 geteuid(),dir);
		return ERR;
	}
	return OK;
};

static int do_user(request_rec *r, t_cfg *cfg)
{
	char user[32];
	char *ptr=NULL;
	uid_t uid;
	gid_t gid;
	struct passwd *pwd=NULL;
	char *chrootdir=NULL;
	char path[PATH_MAX]={'\0'};
	char real_path[PATH_MAX]={'\0'};

	DEBUG("do_user() for uri: %s",r->uri);
	// if dynamic config, get user's home dir
	if(cfg->user->uid==DYNAMIC || 
		cfg->user->gid==DYNAMIC || 
		cfg->user->chrootdir==(char *)DYNAMIC) {
		strncpy(user,r->uri+2,sizeof(user)-1);
		if(ptr=strchr(user,'/'))
			*ptr='\0';
		if((pwd=getpwnam(user))==NULL) {
			LOG_ERROR(
				"Cannot get passwd struct for user %s\n",user);
			return HTTP_NOT_FOUND;
		}
	
	}
	// if dynamic config, then *id as user *id, else as cfg->user_*id
	uid=cfg->user->uid==DYNAMIC?pwd->pw_uid:cfg->user->uid;
	gid=cfg->user->gid==DYNAMIC?pwd->pw_gid:cfg->user->gid;
	// if dynamic chroot, then chrootdir to user's home, 
	// else cfg->user->chrootdir
	chrootdir=cfg->user->chrootdir==(char *)DYNAMIC?
		pwd->pw_dir:cfg->user->chrootdir;
	if(cfg->user->relative!=NULL) {
	// if relative chroot then, we have to get full path for this relative
	// dir by calling realpath()
		ap_snprintf(path,sizeof(path),"%s/%s",
			chrootdir,cfg->user->relative);
		DEBUG("Doing relative chroot to %s",path);
		if(realpath(path,real_path)==NULL) {
			LOG_ERROR("Cannot get realpath for %s",path);
			return HTTP_SERVICE_UNAVAILABLE;
		}
		chrootdir=real_path;
	}
	if(chrootdir!=NULL && try_chroot(r, chrootdir)!=OK)
		return HTTP_SERVICE_UNAVAILABLE;
	if(gid!=NOTHING && try_setgid(r, gid)!=OK)
		return HTTP_SERVICE_UNAVAILABLE;
	// if want call setgroups(0,NULL) - do it
	if(cfg->control->setgroups==1 && try_setgroups(r)!=OK)
		return HTTP_SERVICE_UNAVAILABLE;
	if(uid!=NOTHING && try_setuid(r, uid)!=OK) 
		return HTTP_SERVICE_UNAVAILABLE;
	cfg->control->didall=1;
	return DECLINED;
}

static int do_non_user(request_rec *r, t_cfg *cfg)
{
	DEBUG("do_non_user: chrootdir: %s, gid: %d, uid %d",
		cfg->host->chrootdir,cfg->host->gid,cfg->host->uid);
	if(cfg->host->chrootdir!=NULL
		&& try_chroot(r, cfg->host->chrootdir)!=OK)
		return HTTP_SERVICE_UNAVAILABLE;
	if(cfg->host->gid!=NOTHING && try_setgid(r, cfg->host->gid)!=OK)
		return HTTP_SERVICE_UNAVAILABLE;
	// if want call setgroups(0,NULL) - do it
	if(cfg->control->setgroups==1 && try_setgroups(r)!=OK)
		return HTTP_SERVICE_UNAVAILABLE;
	if(cfg->host->uid!=NOTHING && try_setuid(r, cfg->host->uid)!=OK) 
		return HTTP_SERVICE_UNAVAILABLE;
	cfg->control->didall=1;
	return DECLINED;
}

static int mod_diffprivs_first(request_rec *r)
{
	t_cfg *cfg;

	DEBUG("Called mod_diffprivs_first");
	cfg=ap_get_module_config(r->server->module_config,&diffprivs_module);
	// if everything is done or no directive, then back
	if(cfg->control->didall==1 || cfg->control->pres!=1)
		return DECLINED;
	// don't allow jump visits
		DEBUG("Check jump visit");
		if(last_server_name==NULL && r->hostname)
			last_server_name=strdup(r->hostname);
		// if last server is diffrent than this server or 
		// last user is diffrent than this user, then don't allow
		// for request
		if(r->hostname && strcasecmp(last_server_name,r->hostname)) {
			LOG_NOERROR("Jump visit detected from %s to %s",
				last_server_name,r->hostname);
			//let client know, that this error is really temporary
			//from RFC
			ap_table_setn(r->err_headers_out, "Retry-After", "1");
			//this will actually work in IE/Netscape/Mozilla
			ap_table_setn(r->err_headers_out, "Refresh", "1");
			//BTW, following error closes connection, but child
			//is not required to exit...
			return HTTP_SERVICE_UNAVAILABLE;
		}
		DEBUG("After check jump visit => OK");
	// if ~user request
	if(r->uri[0]=='/' && r->uri[1]=='~')
		// if have to do user work, then ok
		if(cfg->control->user==1)
			return do_user(r,cfg);
		// else return error (no ~user for this server)
		else {
			LOG_NOERROR("~user requests are not allowed for "
				"this server. URI: %s\n",r->uri);
			return HTTP_FORBIDDEN;
		}
	// if not ~user request then do normal work
	else
		return do_non_user(r,cfg);
}

static int parse_id(int isuid, uid_t *id, const char *idname, char *error) {
	char *endptr=NULL;
	struct passwd *pwd=NULL;
	struct group *grp=NULL;

	if(idname==NULL) {
		*id=NOTHING;
		return OK;
	};
	// idname not given
	if(!strcasecmp("DoNothing",idname)) {
		*id=NOTHING;
		return OK;
	};
	// dynamic config
	if(isuid ? !strcasecmp("AsUserUid",idname)
	         : !strcasecmp("AsUserGid",idname)) {
		*id=DYNAMIC;
		return OK;
	};
	// idname given as #id
	if(*idname=='#') {
		*id=strtol(idname+1,&endptr,10);
		// if strtol return error
		if(endptr && *endptr) {
			ap_snprintf(error,ESS,"Invalid ID: %s\n",idname);
			return ERR;
		}
	}
	else {
	    // user/group given as name
	    if (isuid) {
		pwd=getpwnam(idname);
		if(pwd==NULL) {
			ap_snprintf(error,ESS,
				"Cannot get uid for user %s\n",idname);
			return ERR;
		}
		*id=pwd->pw_uid;
		if(*id<MIN_UID) {
		    ap_snprintf(error,ESS,"Miminal allowed uid is %d, but used %d "
				"for %s",MIN_UID,*id,idname);
		    return ERR;
		}
	    } else {
		grp=getgrnam(idname);
		if(grp==NULL) {
			ap_snprintf(error,ESS,
				"Cannot get gid for group %s\n",idname);
			return ERR;
		}
		*id=grp->gr_gid;
		if(*id<MIN_GID) {
		    ap_snprintf(error,ESS,"Miminal allowed gid is %d, but used %d "
				"for %s",MIN_GID,*id,idname);
		    return ERR;
		}
	    };
	}
	return OK;
}

static int parse_dir(char **chrootdir,const char *dir,char *error) {
	// if no chroot
	if(dir==NULL) {
		*chrootdir=NULL;
		return OK;
	}
	// if dynamic config
	if(!strncasecmp("AsUserHome",dir,10)) {
		*chrootdir=(char *)DYNAMIC;
		return OK;
	}
	// if dir not exist
	if(access(dir,F_OK)!=0) {
		ap_snprintf(error,ESS,"File not found: %s\n",dir);
		return ERR;
	}
	// if it's under UNDER_HERE value from config.h
	if(strncmp(UNDER_HERE,dir,strlen(UNDER_HERE))) {
		ap_snprintf(error,ESS,
			"Cannot do chroot for dir %s (it is not under %s)\n",
			dir,UNDER_HERE);
		return ERR;
	}
	// don't want relative paths as directory
	if(strstr(dir,"..")) {
		ap_snprintf(error,ESS,
			"Not allowed characters in path: %s\n",dir);
		return ERR;
	}
	// as above
	if(dir[0]!='/') {
		ap_snprintf(error,ESS,"Path must be non-relative: %s\n",dir);
		return ERR;
	}
	*chrootdir=(char *)dir;
	return OK;
}

// get Privs argument from config
static const char *privs(cmd_parms *cmd, void *mconfig,
                             const char *user,
                             const char *group,
                             const char *dir) {
	static char error[ESS];
	// get config
	t_cfg *cfg = ap_get_module_config(cmd->server->module_config,
						&diffprivs_module);
	memset(error,0,ESS);
	if(parse_id(1, &cfg->host->uid,user,error)!=OK)
		return error;
	if(parse_id(0, &cfg->host->gid,group,error)!=OK)
		return error;
	if(parse_dir(&cfg->host->chrootdir,dir,error)!=OK)
		return error;
	// dynamic config can be only for PrivsUser directive, not here
	if(cfg->host->uid==DYNAMIC || 
		cfg->host->gid==DYNAMIC || 
		cfg->host->chrootdir==(char *)DYNAMIC) {
		ap_snprintf(error,ESS,
			"Cannot do dynamic config in this directive");
		return error;
	}
	cfg->control->pres=1;
	return NULL;
};

static const char *privsuser(cmd_parms *cmd,void *mconfig,
				const char *user,
				const char *group,
				const char *dir)
{
	static char error[ESS];
	// get config
	t_cfg *cfg = ap_get_module_config(cmd->server->module_config,
						&diffprivs_module);
	memset(error,0,ESS);
	if(parse_id(1, &cfg->user->uid,user,error)!=OK)
		return error;
	if(parse_id(0, &cfg->user->gid,group,error)!=OK)
		return error;
	if(parse_dir(&cfg->user->chrootdir,dir,error)!=OK)
		return error;
	// if there is something after AsUserHome then set it as relative
	if(cfg->user->chrootdir==(char *)DYNAMIC && strlen(dir)>10)
		cfg->user->relative=(char *)dir+strlen("AsUserHome");
	cfg->control->user=1;
	cfg->control->pres=1;
	return NULL;
}

static const char *privsgroups(cmd_parms *cmd, void *mconfig, int bool)
{
	// get config
	t_cfg *cfg = ap_get_module_config(cmd->server->module_config,
						&diffprivs_module);
	cfg->control->setgroups=bool;
	cfg->control->pres=1;
	return NULL;
}

// table of commands
static const command_rec mod_diffprivs_commands[] =
{
	{
	"Privs", privs, NULL, RSRC_CONF, TAKE123,
		"Diffrent process privileges: \"Privs uid [gid [chroot_dir]]\""
	},

	{
	"PrivsUser",privsuser,NULL,RSRC_CONF,TAKE123,
		"Diffrent privileges for user's "
			"(www.server.com/~user) requests: \""
			"PrivsUser uid [gid [chroot_dir]]\""
	},
	{
	"PrivsGroups",privsgroups,NULL,RSRC_CONF,FLAG,
		"Call or not setgroups(0,NULL): \"PrivsGroups On|Off\""
	},
     {NULL}
};

module MODULE_VAR_EXPORT diffprivs_module =
{
   STANDARD_MODULE_STUFF,
     mod_diffprivs_init,                // add info about module
     NULL, NULL,
     mod_diffprivs_server_config,       // create server config
     NULL,
     mod_diffprivs_commands,            // module commands in httpd.conf
     NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
     mod_diffprivs_first                // when server read request.
};
